# (c) Copyright 2009-2013. CodeWeavers, Inc.

import os
import traceback

import gobject
gobject.threads_init()
import gtk

import cxlog
import cxutils
import cxobservable
import distversion

import cxguitools
import pyop

import demoutils

# for localization
from cxutils import cxgettext as _

# These are the observable events
DIALOG_CLOSED = 'dialog_closed'


#
# Dialog instance
#

class RegisterDialogController(cxobservable.Object):

    # This is the list of observable events
    observable_events = frozenset((DIALOG_CLOSED,))

    def __init__(self):
        cxobservable.Object.__init__(self)

        # Import widget modules so gtk.Builder() finds them
        import cxhtmltextview # pylint: disable=W0612
        self.xml = gtk.Builder()
        self.xml.set_translation_domain("crossover")
        self.xml.add_from_file(cxguitools.get_ui_path("registerdialog"))
        self.xml.connect_signals(self)

        try:
            icon = cxguitools.get_std_icon('cxregister')
            self.xml.get_object("RegisterDialog").set_icon(icon)
        except gobject.GError:
            cxlog.warn("Couldn't load icon:\n%s" % traceback.format_exc())

        explanationText = _("""<body>
                             <p>To buy a full version of %(productname)s, visit:</p>
                             <p><a href="%(url)s">%(visibleurl)s</a></p>
                             <p>During your purchase you will be asked to create an account using your email
                             address and a password. Enter your email address and password below, and click
                             'Register' to register and unlock your install.</p>
                          </body>""") % {
            'url': distversion.DEMO_URL,
            'visibleurl': distversion.DEMO_URL_VISIBLE,
            'productname': distversion.PRODUCT_NAME}

        self.xml.get_object("ExplanationTextView").display_html(explanationText)

        registerButton = self.xml.get_object("RegisterButton")
        registerButton.set_flags(gtk.CAN_DEFAULT)

        self.normalProgbar = self.xml.get_object("RegistrationProgbar")
        self.normalProgbar.set_pulse_step(0.05)
        self.normalProgbar.hide()

        self.serialProgbar = self.xml.get_object("SerialRegistrationProgbar")
        self.serialProgbar.set_pulse_step(0.05)
        self.serialProgbar.hide()

        self.progBar = self.normalProgbar

        self.normalWindow = self.xml.get_object('RegisterDialog')
        self.serialCodeWindow = self.xml.get_object('SerialCodeRegisterWindow')
        self.mainWindow = self.normalWindow

        self.update_sensitivities(None)

        self.xml.get_object("EmailEntry").grab_focus()


    def update_sensitivities(self, _caller):

        emailEntry = self.xml.get_object("EmailEntry")
        passwordEntry = self.xml.get_object("PasswordEntry")
        registerButton = self.xml.get_object("RegisterButton")

        if len(passwordEntry.get_text()) and len(emailEntry.get_text()):
            registerButton.set_sensitive(True)
        else:
            registerButton.set_sensitive(False)
            registerButton.grab_default()

    def sensitivities_at_start_registration(self):
        emailEntry = self.xml.get_object("EmailEntry")
        passwordEntry = self.xml.get_object("PasswordEntry")
        registerButton = self.xml.get_object("RegisterButton")
        cancelButton = self.xml.get_object("CancelButton")
        serialEmailEntry = self.xml.get_object("serialEmailEntry")
        serialPasswordEntry = self.xml.get_object("serialPasswordEntry")

        registerButton.set_sensitive(False)
        cancelButton.set_sensitive(False)
        passwordEntry.set_sensitive(False)
        emailEntry.set_sensitive(False)
        serialEmailEntry.set_sensitive(False)
        serialPasswordEntry.set_sensitive(False)

        self.animateEvent = gobject.timeout_add(100, self.progbar_pulse)

        self.progBar.show()

    def sensitivities_after_registration(self):
        emailEntry = self.xml.get_object("EmailEntry")
        passwordEntry = self.xml.get_object("PasswordEntry")
        registerButton = self.xml.get_object("RegisterButton")
        cancelButton = self.xml.get_object("CancelButton")
        serialEmailEntry = self.xml.get_object("serialEmailEntry")
        serialPasswordEntry = self.xml.get_object("serialPasswordEntry")

        registerButton.set_sensitive(True)
        cancelButton.set_sensitive(True)
        passwordEntry.set_sensitive(True)
        emailEntry.set_sensitive(True)
        serialEmailEntry.set_sensitive(True)
        serialPasswordEntry.set_sensitive(True)

        gobject.source_remove(self.animateEvent)

        self.progBar.hide()
        self.update_sensitivities(None)

    def progbar_pulse(self):
        self.progBar.pulse()
        return True

    def normal_register_clicked(self, _caller):
        emailEntry = self.xml.get_object("EmailEntry")
        passwordEntry = self.xml.get_object("PasswordEntry")

        self.sensitivities_at_start_registration()
        registerOp = RegisterOperation(self, emailEntry.get_text(), passwordEntry.get_text())
        pyop.sharedOperationQueue.enqueue(registerOp)

    def serial_register_finished(self, success, errorString):

        if success:
            emailEntry = self.xml.get_object("serialEmailEntry")
            passwordEntry = self.xml.get_object("serialPasswordEntry")

            registerOp = RegisterOperation(self,
                                           emailEntry.get_text(),
                                           passwordEntry.get_text())
            pyop.sharedOperationQueue.enqueue(registerOp)
        else:
            self.sensitivities_after_registration()
            cxguitools.CXMessageDlg(primary=_("Registration failed"),
                                    secondary=errorString,
                                    buttons=gtk.BUTTONS_OK,
                                    parent=self.mainWindow,
                                    message_type=gtk.MESSAGE_ERROR)


    def normal_register_finished(self, success, errorString):
        self.sensitivities_after_registration()

        if success:
            cxguitools.CXMessageDlg(_("Registration successful."), buttons=gtk.BUTTONS_OK, response_function=self.quit_requested, parent=self.mainWindow, message_type=gtk.MESSAGE_INFO)
        else:
            cxguitools.CXMessageDlg(primary=_("Registration failed"), secondary=errorString, buttons=gtk.BUTTONS_OK, parent=self.mainWindow, message_type=gtk.MESSAGE_ERROR)

    def get_email_format_error(self, in_text):
        if in_text.count("@") != 1:
            return _("Please enter a valid email address.\n(e.g. myusername@myemailprovider.com)")
        else:
            return ""

    def get_serial_format_error(self, in_text):
        trimString = in_text.replace("-", "")

        if len(trimString) == 19:
            return ""

        if len(trimString) < 19:
            return _("This serial code is not valid - it is too short.\n(Example: XXX-XXXX-XXXX-XXXXXXXX).")
        else:
            return _("This serial code is not valid - it is too long.\n(Example: XXX-XXXX-XXXX-XXXXXXXX).")

    def format_serial_code(self, in_code):
        s = in_code.replace("-", "")

        return s[:3] + "-" + s[3:7] + "-" + s[7:11] + "-" + s[11:]


    def serial_register_clicked(self, _caller):
        emailEntry = self.xml.get_object("serialEmailEntry")
        passwordEntry = self.xml.get_object("serialPasswordEntry")
        nameEntry = self.xml.get_object("serialNameEntry")
        codeEntry = self.xml.get_object("serialCodeEntry")

        emailFormatError = self.get_email_format_error(emailEntry.get_text())
        sCodeFormatError = self.get_serial_format_error(codeEntry.get_text())

        emailErrLabel = self.xml.get_object("emailAddressErrorLabel")
        serialCodeErrLabel = self.xml.get_object("serialCodeErrorLabel")

        emailErrLabel.set_text(emailFormatError)
        serialCodeErrLabel.set_text(sCodeFormatError)

        if not (emailFormatError == "" and sCodeFormatError == ""):
            return

        codeEntry.set_text(self.format_serial_code(codeEntry.get_text()))

        self.sensitivities_at_start_registration()
        registerOp = SerialCodeAccountCreationOperation(self,
                                                        emailEntry.get_text(),
                                                        passwordEntry.get_text(),
                                                        nameEntry.get_text(),
                                                        codeEntry.get_text())
        pyop.sharedOperationQueue.enqueue(registerOp)

    def show_serial_register_window(self, _caller):
        emailErrLabel = self.xml.get_object("emailAddressErrorLabel")
        emailErrLabel.set_text("")
        serialCodeErrLabel = self.xml.get_object("serialCodeErrorLabel")
        serialCodeErrLabel.set_text("")

        self.mainWindow = self.serialCodeWindow
        self.progBar = self.serialProgbar
        self.normalWindow.hide()
        self.present()

    def show_normal_register_window(self, _caller):
        self.mainWindow = self.normalWindow
        self.progBar = self.normalProgbar
        self.serialCodeWindow.hide()
        self.present()

    def present(self):
        self.mainWindow.present()

    def quit_requested(self, _caller):
        self.mainWindow.destroy()
        self.emit_event(DIALOG_CLOSED)
        cxguitools.toplevel_quit()


#
# Global dialog management
#

DIALOG = None

def _dialog_closed(_event=None, _data=None, *_args):
    # pylint: disable=W0603
    global DIALOG
    DIALOG = None

def open_or_show(close_when_finished=False, *args, **kwargs):
    # pylint: disable=W0603

    license_file = os.path.join(cxutils.CX_ROOT, "etc", "license.txt")
    sig_file = os.path.join(cxutils.CX_ROOT, "etc", "license.sig")
    (isdemo, username, _date, _licenseid, _revoked) = demoutils.demo_status(license_file, sig_file)

    if isdemo:
        global DIALOG
        if DIALOG is None:
            DIALOG = RegisterDialogController(*args, **kwargs)
            DIALOG.add_observer(DIALOG_CLOSED, _dialog_closed)
        else:
            DIALOG.present()
        return DIALOG
    else:
        if username is not None:
            msg = _("This install of %(product)s has already been registered by:\n\n%(name)s\n\nIt is fully enabled. Thank you for your support!") % {'product': distversion.PRODUCT_NAME, 'name': username}
        else:
            msg = _("This install of %s is already fully enabled.\n\nThank you for your support!") % distversion.PRODUCT_NAME
        if close_when_finished:
            import sys
            my_response_func = sys.exit
        else:
            my_response_func = None
        cxguitools.CXMessageDlg(msg, response_function=my_response_func, buttons=gtk.BUTTONS_OK, message_type=gtk.MESSAGE_INFO)
        return None


#
# Operations
#

class SerialCodeAccountCreationOperation(pyop.PythonOperation):

    def __init__(self, inRegisterController, inLogin, inPassword,
                 inUserName, inSerialCode):
        pyop.PythonOperation.__init__(self)
        self.registerController = inRegisterController
        self.login = inLogin
        self.password = inPassword
        self.success = False
        self.userName = inUserName
        self.serialCode = inSerialCode
        self.errorString = None

    def __unicode__(self):
        return "SerialCodeAccountCreationOperation with username " + self.login

    def main(self):
        self.success = False

        self.errorString = demoutils.create_acct_apply_scode(self.login,
                                                             self.password,
                                                             self.userName,
                                                             self.serialCode)
        if self.errorString is None:
            self.success = True

        return

    def finish(self):
        self.registerController.serial_register_finished(self.success,
                                                         self.errorString)
        pyop.PythonOperation.finish(self)



class RegisterOperation(pyop.PythonOperation):

    def __init__(self, inRegisterController, inLogin, inPassword):
        pyop.PythonOperation.__init__(self)
        self.registerController = inRegisterController
        self.login = inLogin
        self.password = inPassword
        self.success = False
        self.errorString = None

    def __unicode__(self):
        return "RegisterOperation with username " + self.login

    def main(self):
        self.success = False

        tempLicenseFile, tempSigFile, self.errorString = demoutils.download_and_split_license_file(self.login, self.password)
        if not tempSigFile:
            return

        (isDemo, username, _date, _licenseid, _revoked) = demoutils.demo_status(tempLicenseFile, tempSigFile)
        if isDemo:
            if not username:
                self.errorString = _("Unable to download a valid license file.")
            else:
                self.errorString = _("Your support contract is not valid for this product. You may need to purchase a support extension or an additional license.")
            os.remove(tempLicenseFile)
            os.remove(tempSigFile)
            return

        cmd = []
        cert_dir = os.path.join(cxutils.CX_ROOT, "etc")
        if not demoutils.can_register_as_current_user(cert_dir):
            cmd.extend((os.path.join(cxutils.CX_ROOT, "bin", "cxsu"), "--ignore-home"))
        cmd.extend((os.path.join(cxutils.CX_ROOT, "bin", "cxregister"),
                    "--install", tempLicenseFile, tempSigFile))

        retcode, _out, _err = cxutils.run(cmd, stderr=cxutils.GRAB)
        os.remove(tempLicenseFile)
        os.remove(tempSigFile)
        if retcode != 0:
            self.errorString = _("Registration failed")
        else:
            self.success = True
            self.errorString = None


    def finish(self):
        self.registerController.normal_register_finished(self.success,
                                                         self.errorString)
        pyop.PythonOperation.finish(self)
