# (c) Copyright 2009-2013. CodeWeavers, Inc.

import errno
import socket
import urllib
import urllib2
import traceback
import time

import proxyinfo
import cxlog
import cxutils

proxyinfo.install_default_opener()

_WEEKDAY_NAMES = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
_MONTH_NAMES   = ["Jan", "Feb", "Mar", "Apr", "May", "Jun",
                  "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"]

def format_http_date(date_time):
    """Converts a datetime object to an RFC-1123-compliant string for
    use in HTTP headers. Note that this does not rely on strftime() to avoid
    locale issues.
    """
    return "%s, %02d %s %04d %02d:%02d:%02d UTC" % (
        _WEEKDAY_NAMES[date_time.weekday()],
        date_time.day,
        _MONTH_NAMES[date_time.month - 1],
        date_time.year,
        date_time.hour,
        date_time.minute,
        date_time.second
        )

def _get_rfc822_value(string):
    if string.startswith('"'):
        # quoted-string
        return string.split('"', 2)[1]
    else:
        # token
        def is_token_breaker(c):
            return c in ' ()<>@,;:\\"/[]?=' or not 0x1f < ord(c) < 0x7f
        chars = []
        for c in string:
            if is_token_breaker(c):
                break
            chars.append(c)
        return ''.join(chars)

def url_to_basename(url):
    url = url.split('#', 1)[0].split('?', 1)[0]
    components = url.rsplit('/', 4)
    if components[-1] == 'download' and len(components) > 4 and \
            '.' in components[-2]:
        # URLs of the form '.../foo.exe/download' are quite common
        basename = components[-2]
    else:
        basename = components[-1]
    return urllib.url2pathname(basename)

def url_to_sanitized_basename(url):
    return cxutils.sanitize_windows_filename(url_to_basename(url))

class StopDownload(Exception):
    """Stops the download if raised from a notify_headers or notify_progress
    handler.
    """
    pass

class UrlGetter(object):
    bytes_downloaded = 0
    bytes_total = -1
    last_modified = None
    etag = None
    try_count = 0
    finished = False
    url = ''
    basename = ''
    outfile = None
    infile = None
    notify_progress = None
    notify_finished = None
    notify_failed = None

    blocksize = 4096

    def _do_nothing(self, *args):
        pass

    def __init__(self, url, outfile, notify_progress=_do_nothing, notify_finished=_do_nothing, notify_failed=_do_nothing, notify_headers=_do_nothing, last_modified=None, etag=None):
        # Strip the anchor; it's of no concern to the web server
        self.url = url.split('#', 1)[0]
        self.last_modified = last_modified
        self.etag = etag
        self.basename = url_to_sanitized_basename(url)
        if hasattr(outfile, 'write'):
            self.outfile = outfile
        else:
            self.outfile = open(outfile, 'wb')
        self.notify_headers = notify_headers
        self.notify_progress = notify_progress
        self.notify_finished = notify_finished
        self.notify_failed = notify_failed
        self.timeout = None
        self.userdata = {}

    @staticmethod
    def getfield(headers, name, default=None):
        """HTTPMessage.get() is unusable because when a header is duplicated
        it returns a comma-separated list of the values. This function returns
        the value of the first instance instead."""
        value = headers.getrawheader(name)
        if value is None:
            return default
        return value.strip()

    def _analyze_headers(self, headers):
        self.bytes_total = int(self.getfield(headers, 'content-length', '-1'))
        self.last_modified = self.getfield(headers, 'Last-Modified')
        self.etag = self.getfield(headers, 'ETag')
        self.notify_headers(self, self.infile.geturl(), headers)

        # calculate the basename
        basename = ''
        if 'filename=' in headers.get('content-disposition', ()):
            # FIXME: This is not entirely correct because filename= could appear in some other value
            _unused, filenamevalue = headers['content-disposition'].split('filename=', 1)
            basename = url_to_sanitized_basename(_get_rfc822_value(filenamevalue))
        if not basename:
            # recheck the url in case we were redirected
            basename = url_to_sanitized_basename(self.infile.geturl())
        self.basename = basename

    def _fetch_chunk(self, start_time, start=0, end=-1):
        cxlog.log("fetching url %s from %d to %d" % (cxlog.to_str(self.url),
                                                     start, end))
        req = urllib2.Request(self.url)
        req.add_header('User-Agent', 'Mozilla/5.0 (like Gecko) CrossOver')
        if start != 0 or end != -1:
            byte_range = 'bytes=%d-' % start
            if end != -1:
                byte_range += '%d' % end
            req.add_header('Range', byte_range)
        else:
            if self.last_modified:
                req.add_header("If-Modified-Since", self.last_modified)
            if self.etag:
                req.add_header("If-None-Match", self.etag)

        self.infile = urllib2.urlopen(req, None, self.timeout)
        headers = self.infile.info()
        if self.try_count == 0:
            self._analyze_headers(headers)
        else:
            size =  int(self.getfield(headers, 'content-length', '-1'))
            if size != -1 and size + self.bytes_downloaded != self.bytes_total:
                raise Exception('Download size changed from %d to %d' % (self.bytes_total, size + self.bytes_downloaded))

        self.notify_progress(self)

        while True:
            data = self.infile.read(self.blocksize)
            if data == '':
                break
            else:
                self.outfile.write(data)
                # Not += to avoid ambiguity with the static variable
                self.bytes_downloaded = self.bytes_downloaded + len(data)
                self.notify_progress(self)
        if self.bytes_total != -1 and self.bytes_downloaded != self.bytes_total:
            raise Exception("expected to download %i bytes, got %i" % (self.bytes_total, self.bytes_downloaded))

        # The download is a success
        self.finished = True
        cxlog.log("fetched url %s (%i bytes) in %0.3fs\n" % (cxlog.to_str(self.url), self.bytes_downloaded, time.time() - start_time))
        self.notify_finished(self)

    def fetch(self):
        if self.infile:
            raise Exception("fetch() has already been called")
        start_time = time.time()
        try:
            try:
                last_position = -1
                while not self.finished:
                    last_position = self.bytes_downloaded
                    try:
                        self._fetch_chunk(start_time, self.bytes_downloaded)
                    except socket.error, sockerr:
                        cxlog.log("caught %s last_position=%s bytes_downloaded=%d" % (repr(sockerr), last_position, self.bytes_downloaded))
                        if sockerr.errno != errno.ECONNRESET or \
                                last_position == self.bytes_downloaded or \
                                self.try_count >= 12:
                            # If it's not an ECONNRESET, or we did not make any
                            # progress, or we've tried hard enough already,
                            # then give up
                            self.notify_failed(self, sockerr)
                            raise
                    self.infile.close()
                    # Not += to avoid ambiguity with the static variable
                    self.try_count = self.try_count + 1
                    time.sleep(2)
            except StopDownload:
                return
            except Exception, exception:
                cxlog.log("failed to fetch url %s (%0.3fs):\n%s" % (cxlog.to_str(self.url), time.time() - start_time, traceback.format_exc()))
                self.notify_failed(self, exception)
        finally:
            self.outfile.close()

def main():
    import sys
    def notify_progress(urlgetter):
        sys.stdout.write('\r%d/%d bytes (%.1f%%)' % (urlgetter.bytes_downloaded, urlgetter.bytes_total, 100.0*urlgetter.bytes_downloaded/urlgetter.bytes_total))
        sys.stdout.flush()

    def notify_finished(_urlgetter):
        print '\nSuccess.'

    def notify_failed(_urlgetter, _exception):
        print '\nFailure.'

    getter = UrlGetter(sys.argv[1], sys.argv[2], notify_progress, notify_finished, notify_failed)
    getter.fetch()
    print "Suggested basename:", getter.basename

if __name__ == '__main__':
    main()

